package str.classtree.atermify;

import java.util.*;
import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.Repository;
import org.apache.bcel.Constants;
import org.apache.bcel.util.InstructionFinder;
import aterm.*;
import str.classtree.Flag;
import str.aterm.ATermifier;
import static str.aterm.ATermUtil.*;

public class ATermifyUtil
{
  /** 
   * Class Flags
   */
  public static ATerm buildFlags(JavaClass clazz)
  {
    return buildClassFlags(clazz.getAccessFlags());
  }

  public static ATerm buildClassFlags(int flags)
  {
    return buildFlags(flags, Flag.FOR_CLASS);
  }

  /** 
   * Method Flags
   */
  public static ATerm buildFlags(Method method)
  {
    return buildMethodFlags(method.getAccessFlags());
  }

  private static final HashMap<Integer, ATerm> _method_flag_cache = new HashMap<Integer, ATerm>();
  public static ATerm buildMethodFlags(int flags)
  {
    ATerm result = _method_flag_cache.get(flags);
    if(result == null)
    {
      result = buildFlags(flags, Flag.FOR_METHOD);
      _method_flag_cache.put(flags, result);
    }

    return result;
  }

  /** 
   * Field Flags
   */
  private static final HashMap<Integer, ATerm> _field_flag_cache = new HashMap<Integer, ATerm>();
  public static ATerm buildFieldFlags(int flags)
  {
    ATerm result = _field_flag_cache.get(flags);
    if(result == null)
    {
      result = buildFlags(flags, Flag.FOR_FIELD);
      _field_flag_cache.put(flags, result);
    }

    return result;
  }

  public static ATerm buildFlags(int flags, Set<Flag> from)
  {
    ATermList result = _factory.makeList();
    for(Flag flag : from)
    {
      if(flag.isSetInFlags(flags))
      {
	result = result.insert(flag.toATerm());
      }
    }
    return result;
  }

  private static final HashMap<String, ATerm> _descriptor_cache = new HashMap<String, ATerm>();
  public static ATerm buildMethodDescriptor(String signature)
  {
    ATerm result = _descriptor_cache.get(signature);

    if(result == null)
    {
      result = buildMethodDescriptor(Type.getReturnType(signature), Type.getArgumentTypes(signature));
      _descriptor_cache.put(signature, result);
    }

    return result;
  }

  public static ATerm buildMethodDescriptor(Method method)
  {
    return buildMethodDescriptor(method.getSignature());
  }

  public static ATerm buildMethodDescriptor(Type result, Type[] args)
  {
    ATermList argsATerm = _factory.makeList();
    for(int i = args.length - 1; i >= 0; i--)
    {
      argsATerm = argsATerm.insert(TypeATermifier.SHARED.convert(args[i]));
    }

    return appl("MethodDescriptor", argsATerm, TypeATermifier.SHARED.convert(result));
  }

  private static final HashMap<String, ATerm> _field_descriptor_cache = new HashMap<String, ATerm>();
  public static ATerm buildFieldDescriptor(String signature)
  {
    ATerm result = _descriptor_cache.get(signature);

    if(result == null)
    {
      result = buildFieldDescriptor(Type.getType(signature));
      _descriptor_cache.put(signature, result);
    }

    return result;
  }

  public static ATerm buildFieldDescriptor(Field field)
  {
    return buildFieldDescriptor(field.getSignature());
  }

  public static ATerm buildFieldDescriptor(Type type)
  {
    return appl("FieldDescriptor", TypeATermifier.SHARED.convert(type));
  }


  public static int getInstructionPosition(InstructionList list, InstructionHandle handle)
  {
    InstructionHandle[] lines = list.getInstructionHandles();

    int index = -1;
    for(int i = 0; i < lines.length && index == -1; i++)
    {
      if(lines[i] == handle)
      {
	index = i;
      }
    }

    if(index == -1)
    {
      throw new IllegalArgumentException("InstructionHandle not in list");
    }

    return index;
  }
}
