package str.classtree.atermify;

import java.util.HashMap;
import org.apache.bcel.generic.Type;
import org.apache.bcel.generic.BasicType;
import org.apache.bcel.generic.ReferenceType;
import org.apache.bcel.generic.ArrayType;
import org.apache.bcel.generic.ObjectType;
import aterm.ATerm;
import str.aterm.ATermifier;
import static str.aterm.ATermUtil.*;

public class TypeATermifier implements ATermifier<Type>
{
  public final static TypeATermifier SHARED = new TypeATermifier();

  private HashMap<Type, ATerm> _cache = new HashMap<Type, ATerm>();
  public ATerm convert(Type type)
  {
    ATerm result = _cache.get(type);
    if(result == null)
    {
      result = convertReally(type);
      _cache.put(type, result);
    }

    return result;
  }

  public ATerm convertReally(Type type)
  {
    ATerm result;

    if(type instanceof BasicType)
    {
      String fun = null;
      if(type == Type.BYTE)
	fun = "Byte";
      else if(type == Type.CHAR)
	fun = "Char";
      else if(type == Type.DOUBLE)
        fun = "Double";
      else if(type == Type.FLOAT)
        fun = "Float";
      else if(type == Type.INT)
        fun = "Int";
      else if(type == Type.LONG)
        fun = "Long";
      else if(type == Type.SHORT)
        fun = "Short";
      else if(type == Type.BOOLEAN)
        fun = "Boolean";
      else if(type == Type.VOID)
        fun = "Void";
      else
        throw new IllegalArgumentException("Unknown BaseType: " + type.toString());
      
      result = appl(fun);
    }
    else if(type instanceof ArrayType)
    {
      ArrayType t = (ArrayType) type;
      result = appl("ArrayType", convert(t.getElementType()));
    }
    else if(type instanceof ObjectType)
    {
      ObjectType t = (ObjectType) type;
      result = appl("ObjectType", astring(t.getClassName()));
    }
    else
    {
      throw new IllegalArgumentException("Unknown Type: " + type.toString());
    }

    return result;
  }
}
