#include <aterm2.h>
#include <stdlib.h>
#include <stdbool.h>

struct PosInfo {
  int line;
  int column;
};

/**
 * Constructors
 */
static AFun applFun;
static AFun ambFun;
static AFun startLineColumnFun;
static ATerm optLayout;
static void initPosInfo(void)
{
  applFun = ATmakeAFun("appl", 2, ATfalse);
  ATprotectAFun(applFun);
  ambFun = ATmakeAFun("amb", 1, ATfalse);
  ATprotectAFun(ambFun);
  startLineColumnFun = ATmakeAFun("StartLineColumn", 2, ATfalse);
  ATprotectAFun(startLineColumnFun);

  optLayout = ATparse("cf(opt(layout()))");
  ATprotect(&optLayout);
}

static ATerm traverse(ATerm tree, struct PosInfo* session);

/**
 * Main entry point for Stratego
 */
ATerm dryad_add_pos_info_0_0(void* sl, ATerm t)
{
  initPosInfo();

  struct PosInfo session;
  session.line = 1;
  session.column = 0;

  ATerm top = ATgetArgument(t, 0);
  ATerm whatever = ATgetArgument(t, 1);
  top = traverse(top, &session);
  return (ATerm) ATmakeAppl2(ATmakeAFun("parsetree", 2, ATfalse), top, whatever);
}

static bool isTreeChar(ATerm t)
{
  return ATgetType(t) == AT_INT;
}

static bool isTreeAppl(ATerm t)
{
  return ATgetType(t) == AT_APPL && ATgetAFun(t) == applFun;
}

static bool isTreeLayout(ATerm t)
{
  return isTreeAppl(t) && ATgetArgument(ATgetArgument(t, 0), 1) == optLayout;
}

static ATerm makeTreeAppl(ATerm prod, ATermList args) 
{
  return (ATerm) ATmakeAppl2(applFun, prod, (ATerm) args);
}

static bool isTreeAmb(ATerm t)
{
  return ATgetType(t) == AT_APPL && ATgetAFun(t) == ambFun;
}

static ATerm makeTreeAmb(ATermList t) 
{
  return (ATerm) ATmakeAppl1(ambFun, (ATerm) t);
}

static int getTreeChar(ATerm tree)
{
  return ATgetInt((ATermInt) tree);
}

static ATerm setStartLineColumn(ATerm t, int line, int column)
{
  ATermList annos = (ATermList) ATgetAnnotations(t);
  if(annos == NULL) {
    annos = ATempty;
  }
  annos = ATinsert(annos, (ATerm) ATmakeAppl(startLineColumnFun, ATmakeInt(line), ATmakeInt(column)));
  return ATsetAnnotations(t, (ATerm) annos);
}

static void scan(ATerm t, struct PosInfo* session)
{
  if (isTreeChar(t)) {
    if(getTreeChar(t) == '\n') {
      session->column = 0;
      session->line++;
    }
    else {
      session->column++;
    }
  }
  else if(isTreeAppl(t)) {
    ATermList args = (ATermList) ATgetArgument(t, 1);
    while(args != ATempty) {
      scan(ATgetFirst(args), session);
      args = ATgetNext(args);
    }
  }
  else if (isTreeAmb(t)) {
    scan(ATgetFirst(ATgetArgument(t, 0)), session);
  }
  else {
    ATwarning("unhandled tree: %t\n", t);
  }
}

static ATermList traverseArgs(ATermList args, struct PosInfo* session)
{
  if(args == ATempty) {
    return args;
  }
  else {
    ATerm arg = traverse(ATgetFirst(args), session);
    return ATinsert(traverseArgs(ATgetNext(args), session), arg);
  }
}

static ATerm traverse(ATerm t, struct PosInfo* session)
{
  int line = session->line;
  int column = session->column;

  if (isTreeChar(t)) {
    if (getTreeChar(t) == '\n') {
      session->column = 0;
      session->line++;
    }
    else {
      session->column++;
    }
    return t;
  }
  else if(isTreeLayout(t)) {
    scan(t, session);
    return t;
  }
  else if(isTreeAppl(t)) {
    ATermList args = traverseArgs((ATermList) ATgetArgument(t, 1), session);
    t = makeTreeAppl(ATgetArgument(t, 0), args);
  }
  else if (isTreeAmb(t)) {
    ATermList args = (ATermList) ATgetArgument(t, 0);
    ATermList newArgs = ATmakeList0();
    struct PosInfo startSession = *session;
    struct PosInfo thisSession;

    while(args != ATempty) {
      thisSession = startSession;
      newArgs = ATinsert(newArgs, traverse(ATgetFirst(args), &thisSession));
      args = ATgetNext(args);
    }

    session->line = thisSession.line;
    session->column = thisSession.column;
    t = makeTreeAmb(ATreverse(newArgs));
  }
  else {
    ATwarning("unhandled tree: %t\n", t);
  }

  return setStartLineColumn(t, line, column);
}
