#include <jni.h>
#include <aterm2.h>
#include <srts/stratego.h>
#include <stratego-lib/stratego-lib-common.h>

/**
 * TODO: the JNIEnv must be shared by all virtual machines.
 */

/**
 * Initialization of strjvm
 */
static AFun sym_jvm = 0;

static void (* STRJVM_initialize_next)(void) = NULL;
static void STRJVM_initialize(void)
{
  sym_jvm = ATmakeAFun("sym_jvm", 2, ATfalse);
  ATprotectSymbol(sym_jvm);

  if(STRJVM_initialize_next != NULL)
    STRJVM_initialize_next();
}

static void register_init(void) __attribute__((constructor));
static void register_init(void)
{
  STRJVM_initialize_next  = SRTS_stratego_initialize;
  SRTS_stratego_initialize = &STRJVM_initialize;
}

/**
 * Conversion
 */
static JNIEnv* _STRJVM_env_from_term(ATerm term) {
  JNIEnv* result = NULL;
  if(match_cons(term, sym_jvm)) {
    result = (JNIEnv*) AT_term_to_pointer(ATgetArgument(term, 0));
  }
  return result;
}

static JavaVM* _STRJVM_jvm_from_term(ATerm term) {
  JavaVM* result = NULL;
  if(match_cons(term, sym_jvm)) {
    result = (JavaVM*) AT_term_to_pointer(ATgetArgument(term, 1));
  }
  return result;
}

static ATerm _STRJVM_to_term(JNIEnv* env, JavaVM* jvm) {
  return (ATerm) ATmakeAppl(sym_jvm, AT_pointer_to_term(env), AT_pointer_to_term(jvm));
}

/**
 * Conversion of aterms.
 */
jstring _STRJVM_term_to_string(JNIEnv *env, ATerm string_term) {
  jstring result;
  const char *string = AT_getString(string_term);
  result = (*env)->NewStringUTF(env, string);
  return result;
}

ATerm _STRJVM_read_from_string(JNIEnv *env, jstring string) {
  const char* chars = (*env)->GetStringUTFChars(env, string, NULL);
  ATerm result = ATreadFromString(chars);
  (*env)->ReleaseStringUTFChars(env, string, chars);
  return result;
}

/**
 * Create and destroy a JVM
 */
void _STRJVM_destroy(JNIEnv *env, JavaVM *jvm) {
  if ((*env)->ExceptionOccurred(env)) {
    (*env)->ExceptionDescribe(env);
  }

  (*jvm)->DestroyJavaVM(jvm);
}

ATerm STRJVM_destroy(ATerm term) {
  _STRJVM_destroy(_STRJVM_env_from_term(term), _STRJVM_jvm_from_term(term));
  return (ATerm) ATempty;
}

/**
 * @todo Handle the classpath argument.
 */
ATerm STRJVM_create(ATerm classpath) {
  JNIEnv *env;
  JavaVM *jvm;
  jint res;

  JavaVMInitArgs vm_args;
  vm_args.version = 0x00010004;
  JNI_GetDefaultJavaVMInitArgs(&vm_args);
  vm_args.ignoreUnrecognized = JNI_TRUE;

  JavaVMOption options[2];
  options[0].optionString = AT_getString(classpath);
  options[1].optionString = "-Xss32m";
  vm_args.options = options;
  vm_args.nOptions = 2;
  
  // @todo This produces a warning "dereferencing type-punned pointer will break strict-aliasing rules"
  res = JNI_CreateJavaVM(&jvm, (void**) &env, &vm_args);
  if (res < 0) {
    fprintf(stderr, "strjvm error: Can't create Java VM\n");
    return NULL;
  }

  return _STRJVM_to_term(env, jvm);
}

/**
 * Simple function for testing purposes: invoke static method in class with a string.
 */
ATerm STRJVM_callstatic_string_string(ATerm strjvm, ATerm tclass, ATerm tmethod, ATerm targ) {
  JNIEnv* env = _STRJVM_env_from_term(strjvm);
  jclass class;
  jmethodID mid;
  jstring arg;
  jobject result;

  class = (*env)->FindClass(env, AT_getString(tclass));
  if (class == NULL) {
    fprintf(stderr, "strjvm error: Can't load class\n");
    return NULL;
  }

  mid = (*env)->GetStaticMethodID(env, class, AT_getString(tmethod), "(Ljava/lang/String;)Ljava/lang/String;");
  if (mid == NULL) {
    fprintf(stderr, "strjvm error: Can't get the static method\n");
    return NULL;
  }

  arg = _STRJVM_term_to_string(env, targ);
  result = (*env)->CallStaticObjectMethod(env, class, mid, arg);

  if ((*env)->ExceptionOccurred(env)) {
    (*env)->ExceptionDescribe(env);
    return NULL;
  }

  return _STRJVM_read_from_string(env, (jstring) result);
}

ATerm STRJVM_callstatic_string_stringstring(ATerm strjvm, ATerm tclass, ATerm tmethod, ATerm targ1, ATerm targ2) {
  JNIEnv* env = _STRJVM_env_from_term(strjvm);
  jclass class;
  jmethodID mid;
  jstring arg1;
  jstring arg2;
  jobject result;

  class = (*env)->FindClass(env, AT_getString(tclass));
  if (class == NULL) {
    fprintf(stderr, "strjvm error: Can't load class\n");
    if ((*env)->ExceptionOccurred(env)) {
      (*env)->ExceptionDescribe(env);
    }
    return NULL;
  }

  mid = (*env)->GetStaticMethodID(env, class, AT_getString(tmethod), "(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;");
  if (mid == NULL) {
    fprintf(stderr, "strjvm error: Can't get the static method\n");
    if ((*env)->ExceptionOccurred(env)) {
      (*env)->ExceptionDescribe(env);
    }
    return NULL;
  }

  arg1 = _STRJVM_term_to_string(env, targ1);
  arg2 = _STRJVM_term_to_string(env, targ2);
  result = (*env)->CallStaticObjectMethod(env, class, mid, arg1, arg2);

  if ((*env)->ExceptionOccurred(env)) {
    (*env)->ExceptionDescribe(env);
    return NULL;
  }

  return _STRJVM_read_from_string(env, (jstring) result);
}


